using System.Timers;
using GraphicsManager.Enums;
using GraphicsManager.Interfaces;
using GraphicsManager.Objects.Core;
using OpenTK.Graphics.OpenGL;
using OpenTK.Mathematics;
using OpenTK.Windowing.Common;
using Timer = System.Timers.Timer;

namespace GraphicsManager.Objects;

public class FlowLayout : ParentBase
{
    public FlowLayout(Texture? texture = null)
    :base(texture)
    {
	    AllowHoverFromBehind = true;
		t = new(33);
		t.Enabled = true;
		t.Elapsed += TOnElapsed;
		t.Start();
		Controls.ControlAdded += ControlsOnControlAdded;
		Controls.ControlRemoved += ControlsOnControlRemoved;
	}

	public void ScrollToControl(IRenderObject Control)
	{
		if (Controls.Length < 1 || !Controls.Contains(Control)) throw new Exception("Control does not exist");
		if (Control.Location.Y >= 0 && (Control.Location.Y + Control.Size.Y) <= Size.Y) return;
		int index = Controls.IndexOf(Control);
		bool startbottom = false;
		if (Controls.Length - 1 - index < index) startbottom = true;
		int loc = 0;
		if ((Control.Location.Y + Control.Size.Y) >= Size.Y) loc = Size.Y - Control.Size.Y;
		int start = loc + Control.Size.Y;
		BlockDraw = true;
		if (startbottom)
		{
			for (int i = index + 1; i < Controls.Length; i++)
			{
				start += Controls[i].Size.Y;
			}

			for (int i = Controls.Length - 1; i >= 0; i--)
			{
				start -= Controls[i].Size.Y;
				Controls[i].Location = new(Controls[i].Location.X, start, Controls[i].Location.Z);
			}
		}
		else
		{
			for (int i = index; i >= 0; i--)
			{
				start -= Controls[i].Size.Y;
			}
			for (int i = 0; i < Controls.Length; i++)
			{
				start += Controls[i].Size.Y;
				Controls[i].Location = new(Controls[i].Location.X, start, Controls[i].Location.Z);
			}
		}

		for (int i = Controls.Length - 1; i >= 0; i--)
		{
			if (Controls[i].Location.Y + Controls[i].Size.Y < 0 || Controls[i].Location.Y > Size.Y)
			{
				Controls[i].Visible = false;
			}
		}
		BlockDraw = false;
		if (Parent is not null) Parent.TryDraw();
	}

	public void ScrollToTop()
	{
		if (Controls.Length < 1) return;
		ScrollToControl(Controls[0]);
	}

	public void ScrollToBottom()
	{
		if (Controls.Length < 1) return;
		ScrollToControl(Controls[Controls.Length - 1]);
	}
	
	public override void ReportSizeUpdate(IRenderObject Control)
	{
		if (BlockDraw) return;
		BlockDraw = true;
		bool notfount = true;
		for (int i = 0; i < Controls.Length; i++)
		{
			if (Controls[i] != Control && notfount) continue;
			else
			{
				if (notfount)
				{
					notfount = false;
					continue;
				}

				Controls[i].Location = new(0, Controls[i - 1].Location.Y + Controls[i - 1].Size.Y, Controls[i].Location.Z);
			}
		}

		BlockDraw = false;
	}

	private Task ControlsOnControlRemoved()
	{
		if (Controls.Length < 1) return Task.CompletedTask;
		if (Controls[0].Location.Y > 0) Controls[0].Location = new(0);
		for (int i = 1; i < Controls.Length; i++)
		{
			Controls[i].Location = new(0, Controls[Controls.Length - 1].Location.Y + Controls[Controls.Length - 1].Size.Y, Controls[i].Location.Z);
		}
		return Task.CompletedTask;
	}

	private Task ControlsOnControlAdded(IRenderObject arg)
	{
		if (Controls.Length > 0) arg.Location = new(0, Controls[Controls.Length - 1].Location.Y + Controls[Controls.Length - 1].Size.Y, arg.Location.Z);
		else arg.Location = new(0);
		if (arg is IParent par2)
		{
			for (int i = 0; i < par2.Controls.Length; i++)
			{
				par2.Controls[i].Distance = new(
					par2.Size.X - par2.Controls[i].Size.X - par2.Controls[i].Location.X,
					par2.Size.Y - par2.Controls[i].Size.Y - par2.Controls[i].Location.Y);
				
			}
		}
		arg.Size = new(Size.X, arg.Size.Y);
		if (arg is IParent par)
		{
			for (int i = 0; i < par.Controls.Length; i++)
			{
				bool top = (par.Controls[i].Anchor & ObjectAnchor.Top) == ObjectAnchor.Top;
				bool left = (par.Controls[i].Anchor & ObjectAnchor.Left) == ObjectAnchor.Left;
				bool right = (par.Controls[i].Anchor & ObjectAnchor.Right) == ObjectAnchor.Right;
				bool bottom = (par.Controls[i].Anchor & ObjectAnchor.Bottom) == ObjectAnchor.Bottom;
				if (!top && !bottom) { par.Controls[i].Anchor |= ObjectAnchor.Top; top = true; }
				if (!left && !right) { par.Controls[i].Anchor |= ObjectAnchor.Left; left = true; }
				int lx = (left ? par.Controls[i].Location.X : Size.X - par.Controls[i].Distance.X - par.Controls[i].Size.X);
				int ly = (top ? par.Controls[i].Location.Y : Size.Y - par.Controls[i].Distance.Y - par.Controls[i].Size.Y);
				int sy = (bottom ? Size.Y - par.Controls[i].Distance.Y - ly : par.Controls[i].Size.Y);
				int sx = (right ? Size.X - par.Controls[i].Distance.X - lx : par.Controls[i].Size.X);
				par.Controls[i].Size = new(sx, sy);
				par.Controls[i].Location = new(lx, ly, par.Controls[i].Location.Z);
			}
		}
		arg.Anchor = ObjectAnchor.Left | ObjectAnchor.Right | ObjectAnchor.Top;
		return Task.CompletedTask;
	}
	
	public override Vector2i Size
	{
		get => base.Size;
		set
		{
			BlockDraw = true;
			base.Size = value; 
			for (int i = 0; i < Controls.Length; i++)
			{
				Controls[i].Size = new(value.X, Controls[i].Size.Y);
			}
			ParentResize(new());
			if (Parent is not null) Parent.TryDraw();
			BlockDraw = false;
		}
	}
	
	public Vector2 MousePosition => Window!.MousePosition;

	public override void LoadToParent(IParent Parent, IWindow Window)
	{
		base.LoadToParent(Parent,Window);
		Window.MouseWheel += WindowOnMouseWheel;
	}

	Timer t;
	private Queue<Action> scrols = new();
	private int dist = 0;
	
	private void WindowOnMouseWheel(MouseWheelEventArgs obj)
	{
		try
		{
			if (!MouseInside) return;
			BlockDraw = true;
			dist += (int)obj.OffsetY;
			if (scrols.Any())
			{
				for (int i = 0; i < scrols.Count; i++)
				{
					scrols.Dequeue().Invoke();
				}
				
			}
		}
		catch (Exception e)
		{
			Console.WriteLine(e);
		}
		
	}

	public uint HScrollPixels { get; set; } = 30;
	private void TOnElapsed(object? sender, ElapsedEventArgs e)
	{
		try
		{
			if (BlockDraw)
			{
				scrols.Enqueue(new Action(() =>
				{
					BlockDraw = true;
					int current = dist;
					dist = 0;
					if (Controls.Length < 1) return;
					bool down = current < 0;//scrole wheel dir
					bool moved = false;
					float totoaldist = (current * HScrollPixels);
					if (down && (!Controls[Controls.Length-1].Visible || (Controls[Controls.Length-1].Visible && Controls[Controls.Length-1].Location.Y + Controls[Controls.Length-1].Size.Y > Size.Y))) //put chat up
					{
						moved = true;
						int found = -1;
						bool lfound = false;
						for (int i = 0; i < Controls.Length; i++)
						{
							if (found == -1)
							{
								if (!Controls[i].Visible) continue;
								else
								{
									if ((Controls[i].Location.Y + Controls[i].Size.Y) + totoaldist < 0)
									{
										BlockDraw = true;
										Controls[i].Visible = false;
										continue;
									}

									found = Controls[i].Location.Y + (int)totoaldist;
								}
							}

							if (!lfound)
							{
								BlockDraw = true;
								Controls[i].Location = new(0, found, Controls[i].Location.Z);
								found += Controls[i].Size.Y;

								if (Controls[i].Location.Y <= Size.Y) Controls[i].Visible = true;
								else
								{
									lfound = true;
									Controls[i].Visible = false;
								}
								continue;
							}
							BlockDraw = true;
							if (Controls[i].Visible) Controls[i].Visible = false;
						}
					}
					if (!down && (!Controls[0].Visible || (Controls[0].Visible && Controls[0].Location.Y < 0))) // put chat down
					{
						moved = true;
						int found = -1;
						int lfound = -1;
						for (int i = Controls.Length - 1; i >= 0; i--)
						{
							if (found == -1)
							{
								if (!Controls[i].Visible) continue;
								else
								{
									if (Controls[i].Location.Y + totoaldist > Size.Y)
									{
										BlockDraw = true;
										Controls[i].Visible = false;
										continue;
									}

									found = Controls[i].Location.Y + (int)totoaldist + Controls[i].Size.Y;
								}
							}

							if (lfound == -1)
							{
								BlockDraw = true;
								found -= Controls[i].Size.Y;
								Controls[i].Location = new(0, found, Controls[i].Location.Z);

								if (Controls[i].Location.Y + Controls[i].Size.Y >= 0) Controls[i].Visible = true;
								else
								{
									lfound = i;
									Controls[i].Visible = false;
								}
								continue;
							}
							BlockDraw = true;
							if (Controls[i].Visible) Controls[i].Visible = false;
						}
					}

					if (moved)
					{
						Parent!.TryDraw();
					}

				}));
			}

			BlockDraw = false;
		}
		catch (Exception exception)
		{
			Console.WriteLine(exception);
		}
	}
}