﻿using GraphicsManager.Enums;
using GraphicsManager.Interfaces;
using GraphicsManager.Objects.Core;
using OpenTK.Graphics.OpenGL4;
using OpenTK.Mathematics;
using OpenTK.Windowing.Common;
using OpenTK.Windowing.Common.Input;
using OpenTK.Windowing.Desktop;
using OpenTK.Windowing.GraphicsLibraryFramework;

namespace GraphicsManager.Objects;

public class Rectangle : ITextureObject
{
	public static readonly Dictionary<IGLFWGraphicsContext, Shader> DefaultShader = new();
	public static readonly Dictionary<IGLFWGraphicsContext, Shader> DefaultAlphaShader = new();
	public static readonly Dictionary<IGLFWGraphicsContext, Shader> DefaultAlphaTextureShader = new();

	public ObjectAnchor Anchor { get; set; } = ObjectAnchor.Left | ObjectAnchor.Top;

	public List<Texture> Textures { get; set; } = new();
	public ContextMenu? ContextMenu { get; set; } = null;
	public event Func<string[], Task>? FilesDroped; 

	public Rectangle(Texture? texture)
	{
		if (texture is not null) Textures.Add(texture);;
	}
	public Rectangle()
	{
	}
	
	public virtual void Focus()
	{
		
	}
	public virtual void UnFocus()
	{
		
	}

	private Color4 _BackgroundColor { get; set; } = new(0, 0, 0, 255);

	public Color4 BackgroundColor
	{
		get => _BackgroundColor;
		set
		{
			_BackgroundColor = value;
			if (Parent is not null) Parent.TryDraw();
		}
	}

	private bool _Visible = true;
	public virtual bool Visible
	{
		get => _Visible;
		set
		{
			_Visible = value; 
			if (Parent is not null) Parent.TryDraw();
		}
	}

	public Action? OnDrawAction;
	public Action? OnDrawExitAction;

	public virtual void Draw(int x, int y, int w, int h)
	{
		if (Visible && Loaded)
		{
			if (!Window!.Context.IsCurrent) Window.Context.MakeCurrent();
			foreach (Texture tex in Textures)
			{
				tex.Use();
			}
			Shader.Use();
			if (!Textures.Any() || Shader.Handle == DefaultAlphaShader[Window!.Context].Handle)
			{ 
				GL.Uniform4(0, BackgroundColor);
			}
			GL.BindVertexArray(ArrayObject);
			GL.DrawElements(PrimitiveType.Triangles, Indexs.Length, DrawElementsType.UnsignedShort, 0);
		}
	}
	
	public virtual MouseCursor HoverMouse { get; set; } = MouseCursor.Default;

	public void ForceDistanceUpdate()
	{
		if (Parent is null) return;
		ForceDistanceUpdate(Parent);
	}
	
	public void ForceDistanceUpdate(IParent parent)
	{
		Distance = new(parent.Size.X - Size.X - Location.X, parent.Size.Y - Size.Y - Location.Y);
	}

	public virtual void Clean()
	{
		GL.BindBuffer(BufferTarget.ArrayBuffer, 0);
		GL.DeleteBuffer(BufferObject);
		GL.DeleteBuffer(ElementBufferObject);
		GL.DeleteVertexArray(ArrayObject);
	}

	public virtual void LoadToParent(IParent Parent, IWindow Window)
	{
		if (Loaded) return;
		this.Parent = Parent;
		this.Window = Window;
		if (Shader is null)
		{
			if (!Textures.Any())Shader = DefaultShader[Window.Context];
			else Shader = Texture.TextureShader[Window.Context];
		}
		BufferObject = GL.GenBuffer();
		ArrayObject = GL.GenVertexArray();
		if (Textures.Any() && !Shader.ForTexture)
		{
			Shader = Texture.TextureShader[Window.Context];
		}
		ElementBufferObject = GL.GenBuffer();
		GL.BindBuffer(BufferTarget.ElementArrayBuffer, ElementBufferObject);
		GL.BufferData(BufferTarget.ElementArrayBuffer, Indexs.Length * sizeof(ushort), Indexs, Hint);
		Loaded = true;
		Window.MouseDown += Window_MouseDown;
		Window.FileDrop += WindowOnFileDrop;
		Location = Location;
		if (Distance.X == 0 && Distance.Y == 0) Distance = new(Parent.Size.X - Size.X - Location.X, Parent.Size.Y - Size.Y - Location.Y);
		if (WindowLoaded is not null) WindowLoaded.Invoke(this);
	}

	private void WindowOnFileDrop(FileDropEventArgs obj)
	{
		if (!MouseInside) return;
		if (FilesDroped is not null) _ = FilesDroped.Invoke(obj.FileNames);
	}

	private bool mi = false;

	public bool MouseInside
	{
		get => mi;
		set
		{
			if (Window is null) return;
			if (AllowHoverFromBehind)
			{
				mi = value;
				if (value && MouseEnter is not null)  MouseEnter.Invoke(this);
				if (!value && MouseLeave is not null)  MouseLeave.Invoke(this);
			}
			else
			{
				if (Window.HoveringControl == this && value)
				{
					mi = value;
					if (MouseEnter is not null) MouseEnter.Invoke(this);
				}

				if (Window.HoveringControl != this && !value)
				{
					mi = value;
					if (MouseLeave is not null) MouseLeave.Invoke(this);
				}
			}
		}
	}

	public IParent? Parent { get; private set; }
	public IWindow? Window { get; private set; }
	private bool ahfb = false;

	public bool AllowHoverFromBehind
	{
		get
		{
			return ahfb;
		}
		set
		{
			ahfb = value;
			if (!value)
			{
				if (MouseInside && Window is not null && Window.HoveringControl != this)
				{
					mi = false;
					if (MouseLeave is not null) MouseLeave.Invoke(this);
				}
			}
		}
	}

	private void Window_MouseDown(OpenTK.Windowing.Common.MouseButtonEventArgs e)
	{
		if (MouseInside && e.Button == MouseButton.Button1 && Clicked is not null) _ = Clicked.Invoke(this);
		if (MouseInside && e.Button == MouseButton.Button2 && ContextMenu is not null) ContextMenu.ShowContext(Window!);
		if (!MouseInside && ContextMenu is not null) ContextMenu.HideContext(Window!);
	}

	~Rectangle()
	{
		GL.BindBuffer(BufferTarget.ArrayBuffer, 0);
		GL.DeleteBuffer(BufferObject);
		GL.DeleteBuffer(ElementBufferObject);
		GL.DeleteVertexArray(ArrayObject);
	}

	public Shader Shader { get; set; } = null;
	public int ElementBufferObject { get; private set; }
	public int BufferObject { get; private set; }
	public bool IgnoreHover { get; set; }
	public int ArrayObject { get; private set; }
	private Vector2i size_ = new();
	private Vector3i loc_ = new();

	public float[] Points
	{
		set
		{
			try
			{
				if (Loaded)
				{
					GL.BindBuffer(BufferTarget.ArrayBuffer, BufferObject);
					GL.BindVertexArray(ArrayObject);
					int add = 3;
					if (Textures.Any())
					{
						add = 5;
						GL.EnableVertexAttribArray(Textures.First().Location);
						GL.VertexAttribPointer(Textures.First().Location, 2, VertexAttribPointerType.Float, false, 5 * sizeof(float), 3 * sizeof(float));
						
					}
					GL.VertexAttribPointer(0, 3, VertexAttribPointerType.Float, false, add * sizeof(float), 0);
					GL.EnableVertexAttribArray(0);
					GL.BindBuffer(BufferTarget.ArrayBuffer, BufferObject);
					GL.BufferData(BufferTarget.ArrayBuffer, value.Length * sizeof(float), value, Hint);
					GL.BindVertexArray(ArrayObject);
					GL.BindBuffer(BufferTarget.ElementArrayBuffer, ElementBufferObject);
					GL.BufferData(BufferTarget.ElementArrayBuffer, Indexs.Length * sizeof(ushort), Indexs, Hint);
					if (Window is not null && Window.CanControleUpdate && Loaded) Parent!.TryDraw();
				}
			}
			catch (AccessViolationException v)
			{
				Console.WriteLine(v.Message);
			}
		}
	}

	public ushort[] Indexs { get; set; } = new ushort[6] { 0, 1, 3, 1, 2, 3 };
	public BufferUsageHint Hint { get; set; } = BufferUsageHint.StaticDraw;
	public event Func<IRenderObject, Task>? Clicked;
	public bool Loaded { get; private set; }
	public event Func<IRenderObject, Task>? WindowLoaded;
	public event Func<IRenderObject, Task>? MouseEnter;
	public event Func<IRenderObject, Task>? MouseLeave;
	public object? Tag { get; set; }
	public Vector2i Distance { get; set; }

	private TextureDisplay td = TextureDisplay.Clamped;

	public TextureDisplay TextureDisplay
	{
		get
		{
			return td;
		}
		set
		{
			if (Textures.Count == 0) return;
			TextureDisplay old = td;
			td = value;
			switch (old)
			{
				default:
					switch (value)
					{
						case TextureDisplay.HorizontalCenter or TextureDisplay.ProgressHorizontalCenter:
							float per = (float)Textures[0].RawSize!.Value.Y / Textures[0].RawSize!.Value.X;
							float diff = 0;
							if (Window is not null)
							{
								diff = Window.IntToFloat(Size.Y) + 1;
							}
							Points = new float[]
							{
								saf.X,        laf.Y, 0, Textures[0].MaxText.X,     Textures[0].MaxText.Y,
								saf.X,        saf.Y, 0, Textures[0].MaxText.X,     0,
								laf.X,        saf.Y, 0, 0,                         0,
								laf.X,        laf.Y, 0, 0,                         Textures[0].MaxText.Y,
								saf.X - diff, laf.Y, 0, Textures[0].MaxText.X-per, Textures[0].MaxText.Y,
								saf.X - diff, saf.Y, 0, Textures[0].MaxText.X-per, 0,
								laf.X + diff, saf.Y, 0, per,                       0,
								laf.X + diff, laf.Y, 0, per,                       Textures[0].MaxText.Y,
							};
							Indexs = new ushort[]
							{
								0, 1, 4,
								1, 4, 5,
								4, 5, 6,
								4, 6, 7,
								3, 6, 7,
								2, 3, 6
							};
							break;
					}
					break;
			}
			if (Parent is not null) Parent.TryDraw();
		}
	}
	
	public virtual Vector2i Size
	{
		get
		{
			return size_;
		}
		set
		{
			size_ = value;
			if (Window is null || Parent is null) return;
			Parent.ReportSizeUpdate(this);
			//float[] temp = Points;
			float[] temp = Textures.Count == 0 ? new float[12] : (TextureDisplay == TextureDisplay.Center ? new float[20] : new float[40]);
			saf = new Vector2(Parent.IntToFloat(value.X + loc_.X, false), Parent.IntToFloat(value.Y + loc_.Y, true));
			temp[2] = Location.Z;
			temp[(!Textures.Any() ? 5 : 7)] = Location.Z;
			temp[(!Textures.Any() ? 8 : 12)] = Location.Z;
			temp[(!Textures.Any() ? 11 : 17)] = Location.Z;
			
			temp[0] = saf.X; // top r
			temp[1] = laf.Y;
			temp[(!Textures.Any()? 3 : 5)] = saf.X; // bot r
			temp[(!Textures.Any() ? 4 : 6)] = saf.Y;
			temp[(!Textures.Any() ? 6 : 10)] = laf.X;//bot l
			temp[(!Textures.Any() ? 7 : 11)] = saf.Y;
			temp[(!Textures.Any() ? 9 : 15)] = laf.X;//top l
			temp[(!Textures.Any() ? 10 : 16)] = laf.Y;

			if (Textures.Count > 0)
			{
				Vector2i s = value;
				float diff = Window.IntToFloat(s.Y) + 1;
				float per = (float)Textures[0].RawSize!.Value.Y / Textures[0].RawSize!.Value.X;
				switch (TextureDisplay)
				{
					case TextureDisplay.HorizontalCenter:
						temp[3] = Textures[0].MaxText.X;
						temp[4] = Textures[0].MaxText.Y;
						temp[8] = Textures[0].MaxText.X;
						temp[19] = Textures[0].MaxText.Y;
						temp[20] = saf.X - diff; // top r
						temp[21] = laf.Y;
						temp[22] = 0;
						temp[23] = Textures[0].MaxText.X - per;
						temp[24] = 1;
						
						temp[25] = saf.X - diff; // bot r
						temp[26] = saf.Y;
						temp[27] = 0;
						temp[28] = Textures[0].MaxText.X - per;
						temp[29] = 0;
						
						temp[30] = laf.X + diff; // bot l
						temp[31] = saf.Y;
						temp[32] = 0;
						temp[33] = per;
						temp[34] = 0;
						
						temp[35] = laf.X + diff; // top l
						temp[36] = laf.Y;
						temp[37] = 0;
						temp[38] = per;
						temp[39] = 1;
						break;
					case TextureDisplay.ProgressHorizontalCenter:
						if (s.X > s.Y)
						{
							if (s.X > s.Y + s.Y)
							{
								temp[3] = Textures[0].MaxText.X;
								temp[4] = Textures[0].MaxText.Y;
								temp[8] = Textures[0].MaxText.X;
								temp[19] = Textures[0].MaxText.Y;
								temp[20] = saf.X - diff; // top r
								temp[21] = laf.Y;
								temp[22] = 0;
								temp[23] = Textures[0].MaxText.X - per;
								temp[24] = 1;
						
								temp[25] = saf.X - diff; // bot r
								temp[26] = saf.Y;
								temp[27] = 0;
								temp[28] = 1 - per;
								temp[29] = 0;
						
								temp[30] = laf.X + diff; // bot l
								temp[31] = saf.Y;
								temp[32] = 0;
								temp[33] = per;
								temp[34] = 0;
						
								temp[35] = laf.X + diff; // top l
								temp[36] = laf.Y;
								temp[37] = 0;
								temp[38] = per;
								temp[39] = 1;
							}
							else
							{
								temp[3] = (0.666666f + ((s.X - s.Y) / s.Y)) * Textures[0].MaxText.X;
								temp[4] = (0.666666f + ((s.X - s.Y) / s.Y)) * Textures[0].MaxText.Y;
								temp[8] = (0.666666f + ((s.X - s.Y) / s.Y)) * Textures[0].MaxText.X;
								temp[19] = (0.666666f + ((s.X - s.Y) / s.Y)) * Textures[0].MaxText.Y;
								//start last 33% of texture
								temp[20] =laf.X + diff; // top r
								temp[21] = laf.Y;
								temp[22] = 0;
								temp[23] = per;
								temp[24] = 1;
						
								temp[25] = laf.X + diff; // bot r
								temp[26] = saf.Y;
								temp[27] = 0;
								temp[28] = per;
								temp[29] = 0;
						
								temp[30] = laf.X + diff; // bot l
								temp[31] = saf.Y;
								temp[32] = 0;
								temp[33] = per;
								temp[34] = 0;
						
								temp[35] = laf.X + diff; // top l
								temp[36] = laf.Y;
								temp[37] = 0;
								temp[38] = per;
								temp[39] = 1;
							}
						}
						else
						{
							//first 33% of texture
							temp[3] = (s.X/(s.Y*3));
							temp[4] = Textures[0].MaxText.Y;
							temp[8] = (s.X/(s.Y*3));
							temp[19] = Textures[0].MaxText.Y;
							temp[20] = laf.X; // top r
							temp[21] = laf.Y;
							temp[22] = 0;
							temp[23] = (s.X/(s.Y*3));
							temp[24] = 1;
						
							temp[25] = saf.X; // bot r
							temp[26] = saf.Y;
							temp[27] = 0;
							temp[28] = (s.X/(s.Y*3));
							temp[29] = 0;
						
							temp[30] = saf.X; // bot l
							temp[31] = saf.Y;
							temp[32] = 0;
							temp[33] = (s.X/(s.Y*3));
							temp[34] = 0;
						
							temp[35] = laf.X; // top l
							temp[36] = laf.Y;
							temp[37] = 0;
							temp[38] = (s.X/(s.Y*3));
							temp[39] = 1;
						}
						break;
				}
			}
			
			Points = temp;
		}
	}

	public virtual Vector3i Location
	{
		get
		{
			return loc_;
		}
		set
		{
			loc_ = value;
			if (Window is null || Parent is null) return;
			float[] temp = Textures.Count == 0 ? new float[12] : (TextureDisplay == TextureDisplay.Center ? new float[20] : new float[40]);
			laf = new Vector2(Parent.IntToFloat(value.X, false), Parent.IntToFloat(value.Y, true));
			temp[2] = value.Z;
			temp[(!Textures.Any() ? 5 : 7)] = value.Z;
			temp[(!Textures.Any() ? 8 : 12)] = value.Z;
			temp[(!Textures.Any() ? 11 : 17)] = value.Z;
			
			temp[(!Textures.Any() ? 6 : 10)] = laf.X;
			temp[(!Textures.Any() ? 9 : 15)] = laf.X;
			temp[1] = laf.Y;
			temp[(!Textures.Any() ? 10 : 16)] = laf.Y;
			saf = new Vector2(Parent.IntToFloat(Size.X + value.X, false), Parent.IntToFloat(Size.Y + value.Y, true));
			temp[0] = saf.X;
			temp[(!Textures.Any() ? 3 : 5)] = saf.X;
			temp[(!Textures.Any() ? 4 : 6)] = saf.Y;
			temp[(!Textures.Any() ? 7 : 11)] = saf.Y;
			if (Textures.Count > 0)
			{
				Vector2i s = Size;
				float diff = Window.IntToFloat(s.Y) + 1;
				float per = (float)Textures[0].RawSize!.Value.Y / Textures[0].RawSize!.Value.X;
				switch (TextureDisplay)
				{
					case TextureDisplay.HorizontalCenter:
						temp[3] = Textures[0].MaxText.X;
						temp[4] = Textures[0].MaxText.Y;
						temp[8] = Textures[0].MaxText.X;
						temp[19] = Textures[0].MaxText.Y;
						temp[20] = saf.X - diff; // top r
						temp[21] = laf.Y;
						temp[22] = 0;
						temp[23] = Textures[0].MaxText.X - per;
						temp[24] = 1;
						
						temp[25] = saf.X - diff; // bot r
						temp[26] = saf.Y;
						temp[27] = 0;
						temp[28] = Textures[0].MaxText.X - per;
						temp[29] = 0;
						
						temp[30] = laf.X + diff; // bot l
						temp[31] = saf.Y;
						temp[32] = 0;
						temp[33] = per;
						temp[34] = 0;
						
						temp[35] = laf.X + diff; // top l
						temp[36] = laf.Y;
						temp[37] = 0;
						temp[38] = per;
						temp[39] = 1;
						break;
					case TextureDisplay.ProgressHorizontalCenter:
						if (s.X > s.Y)
						{
							if (s.X > s.Y + s.Y)
							{
								temp[3] = Textures[0].MaxText.X;
								temp[4] = Textures[0].MaxText.Y;
								temp[8] = Textures[0].MaxText.X;
								temp[19] = Textures[0].MaxText.Y;
								temp[20] = saf.X - diff; // top r
								temp[21] = laf.Y;
								temp[22] = 0;
								temp[23] = Textures[0].MaxText.X - per;
								temp[24] = 1;
						
								temp[25] = saf.X - diff; // bot r
								temp[26] = saf.Y;
								temp[27] = 0;
								temp[28] = 1 - per;
								temp[29] = 0;
						
								temp[30] = laf.X + diff; // bot l
								temp[31] = saf.Y;
								temp[32] = 0;
								temp[33] = per;
								temp[34] = 0;
						
								temp[35] = laf.X + diff; // top l
								temp[36] = laf.Y;
								temp[37] = 0;
								temp[38] = per;
								temp[39] = 1;
							}
							else
							{
								temp[3] = (0.666666f + ((s.X - s.Y) / s.Y)) * Textures[0].MaxText.X;
								temp[4] = (0.666666f + ((s.X - s.Y) / s.Y)) * Textures[0].MaxText.Y;
								temp[8] = (0.666666f + ((s.X - s.Y) / s.Y)) * Textures[0].MaxText.X;
								temp[19] = (0.666666f + ((s.X - s.Y) / s.Y)) * Textures[0].MaxText.Y;
								//start last 33% of texture
								temp[20] =laf.X + diff; // top r
								temp[21] = laf.Y;
								temp[22] = 0;
								temp[23] = per;
								temp[24] = 1;
						
								temp[25] = laf.X + diff; // bot r
								temp[26] = saf.Y;
								temp[27] = 0;
								temp[28] = per;
								temp[29] = 0;
						
								temp[30] = laf.X + diff; // bot l
								temp[31] = saf.Y;
								temp[32] = 0;
								temp[33] = per;
								temp[34] = 0;
						
								temp[35] = laf.X + diff; // top l
								temp[36] = laf.Y;
								temp[37] = 0;
								temp[38] = per;
								temp[39] = 1;
							}
						}
						else
						{
							//first 33% of texture
							temp[3] = (s.X/(s.Y*3));
							temp[4] = Textures[0].MaxText.Y;
							temp[8] = (s.X/(s.Y*3));
							temp[19] = Textures[0].MaxText.Y;
							temp[20] = laf.X; // top r
							temp[21] = laf.Y;
							temp[22] = 0;
							temp[23] = (s.X/(s.Y*3));
							temp[24] = 1;
						
							temp[25] = saf.X; // bot r
							temp[26] = saf.Y;
							temp[27] = 0;
							temp[28] = (s.X/(s.Y*3));
							temp[29] = 0;
						
							temp[30] = saf.X; // bot l
							temp[31] = saf.Y;
							temp[32] = 0;
							temp[33] = (s.X/(s.Y*3));
							temp[34] = 0;
						
							temp[35] = laf.X; // top l
							temp[36] = laf.Y;
							temp[37] = 0;
							temp[38] = (s.X/(s.Y*3));
							temp[39] = 1;
						}
						break;
				}
			}

			Points = temp;
			
		}
	}
	
	private Vector2 laf = new(), saf = new();

	public Vector2 LocationAsFloat { get { return laf; } }
	public Vector2 SizeAsFloat { get { return saf; } }
}
